(   Title:  Tests for IEEE 754-2008 reference representation
     File:  ieeefp-ref-test.fs
   Author:  David N. Williams
  Version:  0.5.0
  License:  Public Domain
  Revised:  October 16, 2020

These tests have themselves been tested only on little-endian, 32/64-bit
systems.  They have been tested with double and extended double floats,
but not with single or quad floats.

The revision date may reflect cosmetic changes not logged below.

Version 0.5.0
11Sep20 * Started.
        * Classification tests with nan.
13Sep20 * Added fp environment report, more classification tests.
15Sep20 * Finished classification tests.
16Sep20 * Added T-F< tests.
18Sep20 * Added signed zero T-F< tests.
19Sep20 * Added USE-NATIVE-COMPARISONS switch to test existing
          FLOATING-POINT word set versions.
        * Added:  T-F>  T-F=  T-F<=  T-F>=
27Sep20 * REDEFINED-WARNING-OFF instead of REDEFINED-WARNING off.
30Sep20 * Adapted for revision of ieeefp-ref.fs as a module, removed T-
          prefixes.
 1Oct20 * Removed USE-NATIVE-COMPARISONS switch.
        * Replaced by:  USE-NATIVE  ?USE-NATIVE{  }?USE-NATIVE
        * Added:  FDATUM=  FABSDIFF<  F~  FABS
        * Added logic for selection of ref/native F~ for the testers
          themselves.
 2Oct20 * Added:  FNEGATE  FCOPYSIGN
 3Oct20 * Added:  FMAX  FMIN
 7Oct20 * Added FNEXTUP and FNEXTDOWN tests from fnextup.fs 0.9.0, which
          do not cover gaps for extended doubles.
 9Oct20 * Added nan quieting tests for FNEXTUP and FNEXTDOWN.
11Oct20 * Added:  FSNAN?  FQNAN?
        * Add relaxed versions of FNEXTUP and FNEXTDOWN nan tests using
          FQNAN?.
12Oct20 * Moved RAW-CONSTANT and RAW-DROP tests to rawfloat-test.fs.
        * Replaced negative raw constant definitions by tests of
          RAW-NEGATE.
13Oct20 * Added VERBOSE display about possible fpu conversion of sNaN
          into qNan via F@.  This happens on our legacy macOS 32-bit
          system, but not on our 64-bit Catalina system.
16Oct20 * Added report for integer data endianness.
        * Consolidated tests for <nan> F~ into one section headed by: 
          TESTING F~ nan tolerance
)

\ *** USER CONFIG

\ true CONSTANT USE-TTESTER
false CONSTANT USE-TTESTER

\ true CONSTANT TESTER-USE-HOST-F~
false CONSTANT TESTER-USE-HOST-F~

VARIABLE USE-NATIVE
\ USE-NATIVE off
USE-NATIVE on

(
Turn on the USE-NATIVE variable above to set ?USE-NATIVE{ to hide
reference and reveal host native definitions, and set }?USE-NATIVE to
reveal reference and hide native definitions.  Tests sandwiched between

  ?USE-NATIVE{ ... }?USE-NATIVE

are switched by the value in USE-NATIVE.  Unsandwiched tests always use
reference definitions when they exist in ieeefp-ref.fs.
)

\ *** DEBUGGING

[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ; [THEN]

\ *** INCLUDES AND REPORTS

INCLUDE tester-display.fs
VERBOSE on
INCLUDE ieeefp-ref.fs  \ Loads rawfloat.f, which ensures host has fp.

: ?USE-NATIVE{  ( -- )  USE-NATIVE @ IF previous
  blue-text s" Using native definitions." ?type normal-text ?.cr THEN ;

: }?USE-NATIVE  ( -- )  USE-NATIVE @ IF also IEEEFP-REF
  blue-text s" Using reference definitions" ?type normal-text ?.cr THEN ;

TESTER-USE-HOST-F~ 0= [IF]
also IEEEFP-REF
?.first-cr ?." Tester itself uses reference F~.
[THEN]

USE-TTESTER [IF]
INCLUDE ttester.fs
[ELSE]
INCLUDE ttester-xf.fs
INCLUDE xtester-errors.fs
INCLUDE ftester-errors.fs
[THEN]
TRUE VERBOSE !
REDEFINED-WARNING-OFF
decimal

TESTER-USE-HOST-F~ [IF]
also IEEEFP-REF
?.first-cr ?." Tester itself uses host F~.
[THEN]

?." An address unit is " bits/au . ?." bits.
?." There are " bits/cell . ?." bits/cell.
?." There are " bits/float . ?." bits/float.
LITTLE-ENDIAN?
[IF]   ?." Integer data in memory is little-endian.
[ELSE] ?." Integer data in memory is big-endian.
[THEN]
FLITTLE-ENDIAN?
[IF]   ?." Float data in memory is little-endian.
[ELSE] ?." Float data in memory is big-endian.
[THEN]

:noname  ( -- fp.separate? )  depth >r 1e depth >r fdrop 2r> = ; execute
?." The floating-point stack is "
[IF] ?.( *separate*) [ELSE] ?.( *not separate*) [THEN]
?."  from the data stack.

USE-TTESTER [IF]
(
The ttester default for testing floating-point equality is SET-EXACT,
which is essential here.
)

variable #errors    0 #errors !
:noname  ( c-addr u -- )
(
Display an error message followed by the line that had the error.
)
  1 #errors +! error1 ; error-xt !

: ?.errors  ( -- )  VERBOSE @
  IF blue-text ." #ERRORS: " normal-text #errors @ . THEN ;

[ELSE]  \ USE TTESTER-XF
(
The ttester-xf default for floating-point equality is SET-FT-MODE-EXACT,
which is essential here.
)

: ?.errors  ( -- )  ?.xt-errors s"     " ?type ?.ft-errors ?.cr ;

[THEN]

TESTING FSIGNBIT RAW-NEGATE RAW= |RAW|

t{  1e FSIGNBIT -> false }t
t{ -1e FSIGNBIT -> true }t

t{  1e f>raw raw-negate raw>f -> -1e }t
t{ -1e f>raw raw-negate raw>f ->  1e }t

t{ +inf-raw raw-negate -> -inf-raw }t
t{ 1e 0e f/ f>raw      -> +inf-raw }t

t{ +max-qnan-raw raw-negate -> -max-qnan-raw }t
t{ +min-qnan-raw raw-negate -> -min-qnan-raw }t
t{ +max-snan-raw raw-negate -> -max-snan-raw }t
t{ +min-snan-raw raw-negate -> -min-snan-raw }t
t{ +inf-raw      raw-negate -> -inf-raw }t
t{ +max-n-raw    raw-negate -> -max-n-raw }t
t{ +min-n-raw    raw-negate -> -min-n-raw }t
t{ +max-subn-raw raw-negate -> -max-subn-raw }t
t{ +min-subn-raw raw-negate -> -min-subn-raw }t
t{ +0-raw        raw-negate -> -0-raw }t
 
t{ +max-n-raw    +max-n-raw    raw= -> true }t
t{ +max-n-raw    -max-n-raw    raw= -> false }t
t{ -max-subn-raw -max-subn-raw raw= -> true }t
t{ +min-subn-raw -max-n-raw    raw= -> false }t

t{ +max-qnan-raw +max-qnan-raw raw= -> true }t
t{ -min-snan-raw +inf-raw      raw= -> false }t
t{ +inf-raw      -inf-raw      raw= -> false }t
t{ +0-raw        -0-raw        raw= -> false }t

t{ -1e f>raw |raw| raw>f ->  1e }t
t{  1e f>raw |raw| raw>f ->  1e }t

t{ +min-qnan-raw |raw| -> +min-qnan-raw }t
t{ -min-qnan-raw |raw| -> +min-qnan-raw }t
t{ +0-raw        |raw| -> +0-raw }t
t{ -0-raw        |raw| -> +0-raw }t

TESTING F0= FINFINITY? FNAN? FSNAN? FQNAN? FFINITE? FSUBNORMAL? FNORMAL?

: f0=-raw         ( raw -- [0]? )         raw>f f0= ;
: finf-raw?       ( raw -- [inf]? )       raw>f finfinite? ;
: fnan-raw?       ( raw -- [nan]? )       raw>f fnan? ;
: fsnan-raw?      ( raw -- [snan]? )      raw>f fsnan? ;
: fqnan-raw?      ( raw -- [qnan]? )      raw>f fqnan? ;
: ffin-raw?       ( raw -- [finite]? )    raw>f ffinite? ;
: fsubnormal-raw? ( raw -- [subnormal]? ) raw>f fsubnormal? ;
: fnormal-raw?    ( raw -- [normal]? )    raw>f fnormal? ;

\ If we do this, we should also check for the native nan among the eight
\ "natural" possibilities.
0 [IF]
1e 0e f/ f>raw +inf-raw raw=
[IF]   ?." Native inf is " ?.( +inf)
[ELSE] ?." Native inf is " ?.( -inf)
[THEN] ?." .
[THEN]

t{ +max-qnan-raw f0=-raw -> false }t
t{ +min-qnan-raw f0=-raw -> false }t
t{ +max-snan-raw f0=-raw -> false }t
t{ +min-snan-raw f0=-raw -> false }t
t{ +inf-raw      f0=-raw -> false }t
t{ +max-n-raw    f0=-raw -> false }t
t{ +min-n-raw    f0=-raw -> false }t
t{ +max-subn-raw f0=-raw -> false }t
t{ +min-subn-raw f0=-raw -> false }t
t{ +0-raw        f0=-raw -> true }t

t{ -max-qnan-raw f0=-raw -> false }t
t{ -min-qnan-raw f0=-raw -> false }t
t{ -max-snan-raw f0=-raw -> false }t
t{ -min-snan-raw f0=-raw -> false }t
t{ -inf-raw      f0=-raw -> false }t
t{ -max-n-raw    f0=-raw -> false }t
t{ -min-n-raw    f0=-raw -> false }t
t{ -max-subn-raw f0=-raw -> false }t
t{ -min-subn-raw f0=-raw -> false }t
t{ -0-raw        f0=-raw -> true }t

t{ +max-qnan-raw finf-raw? -> false }t
t{ +min-qnan-raw finf-raw? -> false }t
t{ +max-snan-raw finf-raw? -> false }t
t{ +min-snan-raw finf-raw? -> false }t
t{ +inf-raw      finf-raw? -> true }t
t{ +max-n-raw    finf-raw? -> false }t
t{ +min-n-raw    finf-raw? -> false }t
t{ +max-subn-raw finf-raw? -> false }t
t{ +min-subn-raw finf-raw? -> false }t
t{ +0-raw        finf-raw? -> false }t

t{ -max-qnan-raw finf-raw? -> false }t
t{ -min-qnan-raw finf-raw? -> false }t
t{ -max-snan-raw finf-raw? -> false }t
t{ -min-snan-raw finf-raw? -> false }t
t{ -inf-raw      finf-raw? -> true }t
t{ -max-n-raw    finf-raw? -> false }t
t{ -min-n-raw    finf-raw? -> false }t
t{ -max-subn-raw finf-raw? -> false }t
t{ -min-subn-raw finf-raw? -> false }t
t{ -0-raw        finf-raw? -> false }t

t{  1e        finfinite? -> false }t
t{ -1e        finfinite? -> false }t
t{  1e  0e f/ finfinite? -> true }t
t{ -1e  0e f/ finfinite? -> true }t

t{ +max-qnan-raw fnan-raw? -> true }t
t{ +min-qnan-raw fnan-raw? -> true }t
t{ +max-snan-raw fnan-raw? -> true }t
t{ +min-snan-raw fnan-raw? -> true }t
t{ +inf-raw      fnan-raw? -> false }t
t{ +max-n-raw    fnan-raw? -> false }t
t{ +min-n-raw    fnan-raw? -> false }t
t{ +max-subn-raw fnan-raw? -> false }t
t{ +min-subn-raw fnan-raw? -> false }t
t{ +0-raw        fnan-raw? -> false }t

t{ -max-qnan-raw fnan-raw? -> true }t
t{ -min-qnan-raw fnan-raw? -> true }t
t{ -max-snan-raw fnan-raw? -> true }t
t{ -min-snan-raw fnan-raw? -> true }t
t{ -inf-raw      fnan-raw? -> false }t
t{ -max-n-raw    fnan-raw? -> false }t
t{ -min-n-raw    fnan-raw? -> false }t
t{ -max-subn-raw fnan-raw? -> false }t
t{ -min-subn-raw fnan-raw? -> false }t
t{ -0-raw        fnan-raw? -> false }t

?." If you do not see failed FSNAN? and FQNAN? tests with sNaN
?." inputs, F@ on an sNaN does not convert it to a qNAN.

t{ +max-qnan-raw fsnan-raw? -> false }t
t{ +min-qnan-raw fsnan-raw? -> false }t
t{ +max-snan-raw fsnan-raw? -> true }t \ Quieted by fpu?
t{ +min-snan-raw fsnan-raw? -> true }t \ Quieted by fpu?
t{ +inf-raw      fsnan-raw? -> false }t
t{ +max-n-raw    fsnan-raw? -> false }t
t{ +min-n-raw    fsnan-raw? -> false }t
t{ +max-subn-raw fsnan-raw? -> false }t
t{ +min-subn-raw fsnan-raw? -> false }t
t{ +0-raw        fsnan-raw? -> false }t

t{ -max-qnan-raw fsnan-raw? -> false }t
t{ -min-qnan-raw fsnan-raw? -> false }t
t{ -max-snan-raw fsnan-raw? -> true }t \ Quieted by fpu?
t{ -min-snan-raw fsnan-raw? -> true }t \ Quieted by fpu?
t{ -inf-raw      fsnan-raw? -> false }t
t{ -max-n-raw    fsnan-raw? -> false }t
t{ -min-n-raw    fsnan-raw? -> false }t
t{ -max-subn-raw fsnan-raw? -> false }t
t{ -min-subn-raw fsnan-raw? -> false }t
t{ -0-raw        fsnan-raw? -> false }t

t{ +max-qnan-raw fqnan-raw? -> true }t
t{ +min-qnan-raw fqnan-raw? -> true }t
t{ +max-snan-raw fqnan-raw? -> false }t \ Quieted by fpu?
t{ +min-snan-raw fqnan-raw? -> false }t \ Quieted by fpu?
t{ +inf-raw      fqnan-raw? -> false }t
t{ +max-n-raw    fqnan-raw? -> false }t
t{ +min-n-raw    fqnan-raw? -> false }t
t{ +max-subn-raw fqnan-raw? -> false }t
t{ +min-subn-raw fqnan-raw? -> false }t
t{ +0-raw        fqnan-raw? -> false }t

t{ -max-qnan-raw fqnan-raw? -> true }t
t{ -min-qnan-raw fqnan-raw? -> true }t
t{ -max-snan-raw fqnan-raw? -> false }t \ Quieted by fpu?
t{ -min-snan-raw fqnan-raw? -> false }t \ Quieted by fpu?
t{ -inf-raw      fqnan-raw? -> false }t
t{ -max-n-raw    fqnan-raw? -> false }t
t{ -min-n-raw    fqnan-raw? -> false }t
t{ -max-subn-raw fqnan-raw? -> false }t
t{ -min-subn-raw fqnan-raw? -> false }t
t{ -0-raw        fqnan-raw? -> false }t

t{  1e                fqnan? -> false }t
t{ -1e                fqnan? -> false }t
t{  0e  0e f/         fqnan? -> true }t
t{  0e  0e f/ fnegate fqnan? -> true }t  \ fnegate is tested later

t{ +max-qnan-raw ffin-raw? -> false }t
t{ +min-qnan-raw ffin-raw? -> false }t
t{ +max-snan-raw ffin-raw? -> false }t
t{ +min-snan-raw ffin-raw? -> false }t
t{ +inf-raw      ffin-raw? -> false }t
t{ +max-n-raw    ffin-raw? -> true }t
t{ +min-n-raw    ffin-raw? -> true }t
t{ +max-subn-raw ffin-raw? -> true }t
t{ +min-subn-raw ffin-raw? -> true }t
t{ +0-raw        ffin-raw? -> true }t

t{ -max-qnan-raw ffin-raw? -> false }t
t{ -min-qnan-raw ffin-raw? -> false }t
t{ -max-snan-raw ffin-raw? -> false }t
t{ -min-snan-raw ffin-raw? -> false }t
t{ -inf-raw      ffin-raw? -> false }t
t{ -max-n-raw    ffin-raw? -> true }t
t{ -min-n-raw    ffin-raw? -> true }t
t{ -max-subn-raw ffin-raw? -> true }t
t{ -min-subn-raw ffin-raw? -> true }t
t{ -0-raw        ffin-raw? -> true }t

t{ +max-qnan-raw fsubnormal-raw? -> false }t
t{ +min-qnan-raw fsubnormal-raw? -> false }t
t{ +max-snan-raw fsubnormal-raw? -> false }t
t{ +min-snan-raw fsubnormal-raw? -> false }t
t{ +inf-raw      fsubnormal-raw? -> false }t
t{ +max-n-raw    fsubnormal-raw? -> false }t
t{ +min-n-raw    fsubnormal-raw? -> false }t
t{ +max-subn-raw fsubnormal-raw? -> true }t
t{ +min-subn-raw fsubnormal-raw? -> true }t
t{ +0-raw        fsubnormal-raw? -> false }t

t{ -max-qnan-raw fsubnormal-raw? -> false }t
t{ -min-qnan-raw fsubnormal-raw? -> false }t
t{ -max-snan-raw fsubnormal-raw? -> false }t
t{ -min-snan-raw fsubnormal-raw? -> false }t
t{ -inf-raw      fsubnormal-raw? -> false }t
t{ -max-n-raw    fsubnormal-raw? -> false }t
t{ -min-n-raw    fsubnormal-raw? -> false }t
t{ -max-subn-raw fsubnormal-raw? -> true }t
t{ -min-subn-raw fsubnormal-raw? -> true }t
t{ -0-raw        fsubnormal-raw? -> false }t

t{ +max-qnan-raw fnormal-raw? -> false }t
t{ +min-qnan-raw fnormal-raw? -> false }t
t{ +max-snan-raw fnormal-raw? -> false }t
t{ +min-snan-raw fnormal-raw? -> false }t
t{ +inf-raw      fnormal-raw? -> false }t
t{ +max-n-raw    fnormal-raw? -> true }t
t{ +min-n-raw    fnormal-raw? -> true }t
t{ +max-subn-raw fnormal-raw? -> false }t
t{ +min-subn-raw fnormal-raw? -> false }t
t{ +0-raw        fnormal-raw? -> false }t

\ *** FCONSTANTS

+min-qnan-raw raw>f FCONSTANT +tqnan
-min-qnan-raw raw>f FCONSTANT -tqnan
+min-snan-raw raw>f FCONSTANT +tsnan
-min-snan-raw raw>f FCONSTANT -tsnan
+inf-raw      raw>f FCONSTANT +tinf
-inf-raw      raw>f FCONSTANT -tinf
+max-n-raw    raw>f FCONSTANT +tmax-n
-max-n-raw    raw>f FCONSTANT -tmax-n
+min-n-raw    raw>f FCONSTANT +tmin-n
-min-n-raw    raw>f FCONSTANT -tmin-n

+max-subn-raw raw>f FCONSTANT +tmax-subn
-max-subn-raw raw>f FCONSTANT -tmax-subn
+min-subn-raw raw>f FCONSTANT +tmin-subn
-min-subn-raw raw>f FCONSTANT -tmin-subn
+0-raw        raw>f FCONSTANT +t0
-0-raw        raw>f FCONSTANT -t0

TESTING FABS FNEGATE FCOPYSIGN

?USE-NATIVE{
t{ -1e    FABS ->  1e }t
t{ -t0    FABS -> +t0 }t
t{ -tinf  FABS -> +tinf }t
t{ -tqnan FABS -> +tqnan }t

t{  1e    FABS ->  1e }t
t{ +t0    FABS -> +t0 }t
t{ +tinf  FABS -> +tinf }t
t{ +tqnan FABS -> +tqnan }t

t{ -1e    FNEGATE ->  1e }t
t{ -t0    FNEGATE -> +t0 }t
t{ -tinf  FNEGATE -> +tinf }t
t{ -tqnan FNEGATE -> +tqnan }t

t{  1e    FNEGATE ->  -1e }t
t{ +t0    FNEGATE -> -t0 }t
t{ +tinf  FNEGATE -> -tinf }t
t{ +tqnan FNEGATE -> -tqnan }t
}?USE-NATIVE

t{  1e     1e FCOPYSIGN ->  1e }t
t{ +t0     1e FCOPYSIGN -> +t0 }t
t{ +tinf   1e FCOPYSIGN -> +tinf }t
t{ +tqnan  1e FCOPYSIGN -> +tqnan }t

t{ -1e     1e FCOPYSIGN ->  1e }t
t{ -t0     1e FCOPYSIGN -> +t0 }t
t{ -tinf   1e FCOPYSIGN -> +tinf }t
t{ -tqnan  1e FCOPYSIGN -> +tqnan }t

t{  1e    -1e FCOPYSIGN -> -1e }t
t{ +t0    -1e FCOPYSIGN -> -t0 }t
t{ +tinf  -1e FCOPYSIGN -> -tinf }t
t{ +tqnan -1e FCOPYSIGN -> -tqnan }t

t{ -1e    -1e FCOPYSIGN -> -1e }t
t{ -t0    -1e FCOPYSIGN -> -t0 }t
t{ -tinf  -1e FCOPYSIGN -> -tinf }t
t{ -tqnan -1e FCOPYSIGN -> -tqnan }t

t{  1e    +t0 FCOPYSIGN ->  1e }t
t{ +t0    +t0 FCOPYSIGN -> +t0 }t
t{ +tinf  +t0 FCOPYSIGN -> +tinf }t
t{ +tqnan +t0 FCOPYSIGN -> +tqnan }t

t{ -1e    +t0 FCOPYSIGN ->  1e }t
t{ -t0    +t0 FCOPYSIGN -> +t0 }t
t{ -tinf  +t0 FCOPYSIGN -> +tinf }t
t{ -tqnan +t0 FCOPYSIGN -> +tqnan }t

t{  1e    -t0 FCOPYSIGN -> -1e }t
t{ +t0    -t0 FCOPYSIGN -> -t0 }t
t{ +tinf  -t0 FCOPYSIGN -> -tinf }t
t{ +tqnan -t0 FCOPYSIGN -> -tqnan }t

t{ -1e    -t0 FCOPYSIGN -> -1e }t
t{ -t0    -t0 FCOPYSIGN -> -t0 }t
t{ -tinf  -t0 FCOPYSIGN -> -tinf }t
t{ -tqnan -t0 FCOPYSIGN -> -tqnan }t

t{  1e    +tinf FCOPYSIGN ->  1e }t
t{ +t0    +tinf FCOPYSIGN -> +t0 }t
t{ +tinf  +tinf FCOPYSIGN -> +tinf }t
t{ +tqnan +tinf FCOPYSIGN -> +tqnan }t

t{ -1e    +tinf FCOPYSIGN ->  1e }t
t{ -t0    +tinf FCOPYSIGN -> +t0 }t
t{ -tinf  +tinf FCOPYSIGN -> +tinf }t
t{ -tqnan +tinf FCOPYSIGN -> +tqnan }t

t{  1e    -tinf FCOPYSIGN -> -1e }t
t{ +t0    -tinf FCOPYSIGN -> -t0 }t
t{ +tinf  -tinf FCOPYSIGN -> -tinf }t
t{ +tqnan -tinf FCOPYSIGN -> -tqnan }t

t{ -1e    -tinf FCOPYSIGN -> -1e }t
t{ -t0    -tinf FCOPYSIGN -> -t0 }t
t{ -tinf  -tinf FCOPYSIGN -> -tinf }t
t{ -tqnan -tinf FCOPYSIGN -> -tqnan }t

t{  1e    +tqnan FCOPYSIGN ->  1e }t
t{ +t0    +tqnan FCOPYSIGN -> +t0 }t
t{ +tinf  +tqnan FCOPYSIGN -> +tinf }t
t{ +tqnan +tqnan FCOPYSIGN -> +tqnan }t

t{ -1e    +tqnan FCOPYSIGN ->  1e }t
t{ -t0    +tqnan FCOPYSIGN -> +t0 }t
t{ -tinf  +tqnan FCOPYSIGN -> +tinf }t
t{ -tqnan +tqnan FCOPYSIGN -> +tqnan }t

t{  1e    -tqnan FCOPYSIGN -> -1e }t
t{ +t0    -tqnan FCOPYSIGN -> -t0 }t
t{ +tinf  -tqnan FCOPYSIGN -> -tinf }t
t{ +tqnan -tqnan FCOPYSIGN -> -tqnan }t

t{ -1e    -tqnan FCOPYSIGN -> -1e }t
t{ -t0    -tqnan FCOPYSIGN -> -t0 }t
t{ -tinf  -tqnan FCOPYSIGN -> -tinf }t
t{ -tqnan -tqnan FCOPYSIGN -> -tqnan }t


?USE-NATIVE{
TESTING F<
(
Add corner tests for <finite> <finite> F< ?

The following uses only one of the nans.  It assumes that nan appears in
the reference implementation of the comparisons only through FNAN?, and
that FNAN? is correct.
)

t{  17e   17e  F< -> false }t
t{ -17e  -17e  F< -> false }t

t{ +tqnan +tqnan F< -> false }t
t{ +tinf  +tinf  F< -> false }t
t{ -tinf  -tinf  F< -> false }t

t{ +tqnan +tinf F< -> false }t
t{ +tqnan -tinf F< -> false }t
t{ +tqnan  17e  F< -> false }t
t{ -tinf  +tinf F< -> true  }t
t{  17e   +tinf F< -> true  }t
t{ -17e   -tinf F< -> false }t

t{ +tinf +tqnan F< -> false }t
t{ -tinf +tqnan F< -> false }t
t{  17e  +tqnan F< -> false }t
t{ +tinf -tinf  F< -> false }t
t{ +tinf  17e   F< -> false }t
t{ -tinf -17e   F< -> true  }t

t{ -t0 -t0 F< -> false }t
t{ -t0 -t0 F< -> false }t
t{ -t0 -t0 F< -> false }t
t{ -t0 -t0 F< -> false }t

t{ -t0  +17e F< -> true }t
t{ -t0  +17e F< -> true }t
t{ -t0  -17e F< -> false }t
t{ -t0  -17e F< -> false }t
t{ +17e -t0  F< -> false }t
t{ -17e -t0  F< -> true }t

t{ -t0    +tqnan F< -> false }t
t{ -t0    +tqnan F< -> false }t
t{ +tqnan -t0    F< -> false }t
t{ +tqnan -t0    F< -> false }t

t{ -t0   +tinf F< -> true  }t
t{ -t0   +tinf F< -> true  }t
t{ +tinf -t0   F< -> false }t
t{ +tinf -t0   F< -> false }t

t{ -t0   -tinf F< -> false }t
t{ -t0   -tinf F< -> false }t
t{ -tinf -t0   F< -> true  }t
t{ -tinf -t0   F< -> true  }t

TESTING F>
(
The following is independent of the F< tests.  It uses the same inputs
but replaces the goals.  If F> is implemented by FSWAP F<, the theorem
that they are equivalent is also tested.
)

t{  17e  17e F> -> false }t
t{ -17e -17e F> -> false }t

t{ +tqnan +tqnan F> -> false }t
t{ +tinf  +tinf  F> -> false }t
t{ -tinf  -tinf  F> -> false }t

t{ +tqnan +tinf F> -> false }t
t{ +tqnan -tinf F> -> false }t
t{ +tqnan  17e  F> -> false }t
t{ -tinf  +tinf F> -> false }t
t{  17e   +tinf F> -> false }t
t{ -17e   -tinf F> -> true  }t

t{ +tinf +tqnan F> -> false }t
t{ -tinf +tqnan F> -> false }t
t{  17e  +tqnan F> -> false }t
t{ +tinf -tinf  F> -> true  }t
t{ +tinf  17e   F> -> true  }t
t{ -tinf -17e   F> -> false }t

t{ -t0 -t0 F> -> false }t
t{ -t0 -t0 F> -> false }t
t{ -t0 -t0 F> -> false }t
t{ -t0 -t0 F> -> false }t

t{ -t0  +17e  F> -> false }t
t{ -t0  +17e  F> -> false }t
t{ -t0  -17e  F> -> true  }t
t{ -t0  -17e  F> -> true  }t
t{ +17e -t0   F> -> true  }t
t{ -17e -t0   F> -> false }t

t{ -t0    +tqnan F> -> false }t
t{ -t0    +tqnan F> -> false }t
t{ +tqnan -t0    F> -> false }t
t{ +tqnan -t0    F> -> false }t

t{ -t0   +tinf F> -> false }t
t{ -t0   +tinf F> -> false }t
t{ +tinf -t0   F> -> true  }t
t{ +tinf -t0   F> -> true }t

t{ -t0   -tinf F> -> true  }t
t{ -t0   -tinf F> -> true  }t
t{ -tinf -t0   F> -> false }t
t{ -tinf -t0   F> -> false }t

TESTING F=

t{  17e  17e F= -> true  }t
t{ -17e -17e F= -> true  }t

t{ +tqnan +tqnan F= -> false }t
t{ +tinf  +tinf  F= -> true }t
t{ -tinf  -tinf  F= -> true }t

t{ +tqnan +tinf F= -> false }t
t{ +tqnan -tinf F= -> false }t
t{ +tqnan  17e  F= -> false }t
t{ -tinf  +tinf F= -> false }t
t{  17e   +tinf F= -> false }t
t{ -17e   -tinf F= -> false }t

t{ +tinf +tqnan F= -> false }t
t{ -tinf +tqnan F= -> false }t
t{  17e  +tqnan F= -> false }t
t{ +tinf -tinf  F= -> false }t
t{ +tinf  17e   F= -> false }t
t{ -tinf -17e   F= -> false }t

t{ -t0 -t0 F= -> true  }t
t{ -t0 -t0 F= -> true  }t
t{ -t0 -t0 F= -> true  }t
t{ -t0 -t0 F= -> true  }t

t{ -t0  +17e F= -> false }t
t{ -t0  +17e F= -> false }t
t{ -t0  -17e F= -> false }t
t{ -t0  -17e F= -> false }t
t{ +17e -t0  F= -> false }t
t{ -17e -t0  F= -> false }t

t{ -t0    +tqnan F= -> false }t
t{ -t0    +tqnan F= -> false }t
t{ +tqnan -t0    F= -> false }t
t{ +tqnan -t0    F= -> false }t

t{ -t0   +tinf F= -> false }t
t{ -t0   +tinf F= -> false }t
t{ +tinf -t0   F= -> false }t
t{ +tinf -t0   F= -> false }t

t{ -t0  -tinf F= -> false }t
t{ -t0  -tinf F= -> false }t
t{ -tinf -t0  F= -> false }t
t{ -tinf -t0  F= -> false }t

TESTING F<=

t{  17e  17e F<= -> true  }t
t{ -17e -17e F<= -> true  }t

t{ +tqnan +tqnan F<= -> false }t
t{ +tinf +tinf F<= -> true }t
t{ -tinf -tinf F<= -> true }t

t{ +tqnan +tinf F<= -> false }t
t{ +tqnan -tinf F<= -> false }t
t{ +tqnan  17e  F<= -> false }t
t{ -tinf  +tinf F<= -> true  }t
t{  17e   +tinf F<= -> true  }t
t{ -17e   -tinf F<= -> false }t

t{ +tinf +tqnan F<= -> false }t
t{ -tinf +tqnan F<= -> false }t
t{  17e  +tqnan F<= -> false }t
t{ +tinf -tinf  F<= -> false }t
t{ +tinf  17e   F<= -> false }t
t{ -tinf -17e   F<= -> true  }t

t{ -t0 -t0 F<= -> true  }t
t{ -t0 -t0 F<= -> true  }t
t{ -t0 -t0 F<= -> true  }t
t{ -t0 -t0 F<= -> true  }t

t{ -t0  +17e F<= -> true  }t
t{ -t0  +17e F<= -> true  }t
t{ -t0  -17e F<= -> false }t
t{ -t0  -17e F<= -> false }t
t{ +17e -t0  F<= -> false }t
t{ -17e -t0  F<= -> true  }t

t{ -t0    +tqnan F<= -> false }t
t{ -t0    +tqnan F<= -> false }t
t{ +tqnan -t0    F<= -> false }t
t{ +tqnan -t0    F<= -> false }t

t{ -t0   +tinf F<= -> true  }t
t{ -t0   +tinf F<= -> true  }t
t{ +tinf -t0   F<= -> false }t
t{ +tinf -t0   F<= -> false }t

t{ -t0   -tinf F<= -> false }t
t{ -t0   -tinf F<= -> false }t
t{ -tinf -t0   F<= -> true  }t
t{ -tinf -t0   F<= -> true  }t

TESTING F>=

t{  17e  17e F>= -> true  }t
t{ -17e -17e F>= -> true  }t

t{ +tqnan +tqnan F>= -> false }t
t{ +tinf  +tinf  F>= -> true  }t
t{ -tinf  -tinf  F>= -> true  }t

t{ +tqnan +tinf F>= -> false }t
t{ +tqnan -tinf F>= -> false }t
t{ +tqnan  17e  F>= -> false }t
t{ -tinf +tinf F>= -> false }t
t{  17e  +tinf F>= -> false }t
t{ -17e  -tinf F>= -> true  }t

t{ +tinf +tqnan F>= -> false }t
t{ -tinf +tqnan F>= -> false }t
t{  17e  +tqnan F>= -> false }t
t{ +tinf -tinf  F>= -> true  }t
t{ +tinf  17e   F>= -> true  }t
t{ -tinf -17e   F>= -> false }t

t{ -t0 -t0 F>= -> true  }t
t{ -t0 -t0 F>= -> true  }t
t{ -t0 -t0 F>= -> true  }t
t{ -t0 -t0 F>= -> true  }t

t{ -t0  +17e F>= -> false }t
t{ -t0  +17e F>= -> false }t
t{ -t0  -17e F>= -> true  }t
t{ -t0  -17e F>= -> true  }t
t{ +17e -t0  F>= -> true  }t
t{ -17e -t0  F>= -> false }t

t{ -t0     +tqnan F>= -> false }t
t{ -t0     +tqnan F>= -> false }t
t{  +tqnan -t0    F>= -> false }t
t{  +tqnan -t0    F>= -> false }t

t{ -t0   +tinf F>= -> false }t
t{ -t0   +tinf F>= -> false }t
t{ +tinf -t0   F>= -> true  }t
t{ +tinf -t0   F>= -> true }t

t{ -t0   -tinf F>= -> true  }t
t{ -t0   -tinf F>= -> true  }t
t{ -tinf -t0   F>= -> false }t
t{ -tinf -t0   F>= -> false }t

}?USE-NATIVE

TESTING FDATUM= FABSDIFF<

t{ 1e 0e FDATUM= -> false }t
t{ 0e 1e FDATUM= -> false }t
t{ 1e 1e FDATUM= -> true }t

t{ +t0 +t0 FDATUM= -> true }t
t{ +t0 -t0 FDATUM= -> false }t
t{ -t0 +t0 FDATUM= -> false }t
t{ -t0 -t0 FDATUM= -> true }t

t{ +tinf +tinf FDATUM= -> true }t
t{ +tinf -tinf FDATUM= -> false }t
t{ -tinf +tinf FDATUM= -> false }t
t{ -tinf -tinf FDATUM= -> true }t

t{ +tqnan +tqnan FDATUM= -> true }t
t{ +tqnan -tqnan FDATUM= -> false }t
t{ -tqnan +tqnan FDATUM= -> false }t
t{ -tqnan -tqnan FDATUM= -> true }t

t{ +tmin-subn +t0 FDATUM= -> false }t

t{ 0e 1e  0.5e   FABSDIFF< -> false }t
t{ 0e 1e -0.5e   FABSDIFF< -> false }t
t{ 0e 1e  1.001e FABSDIFF< -> true }t
t{ 0e 1e -1.001e FABSDIFF< -> false }t
t{ 1e 0e  1.001e FABSDIFF< -> true }t
t{ 1e 0e -1.001e FABSDIFF< -> false }t

t{  0e     1e    -tqnan  FABSDIFF< -> false }t
t{  0e    -tqnan  1.001e FABSDIFF< -> false }t
t{ -tqnan  1e     1.001e FABSDIFF< -> false }t

t{ +tqnan  1e    +tqnan  FABSDIFF< -> false }t
t{  1e    +tqnan +tqnan  FABSDIFF< -> false }t
t{ +tqnan -tqnan  1.001e FABSDIFF< -> false }t

t{ +tqnan -tqnan +tqnan FABSDIFF< -> false }t

t{  0e    1e   +tinf   FABSDIFF< -> true }t
t{  0e    1e   -tinf   FABSDIFF< -> false }t
t{  0e   -tinf  1.001e FABSDIFF< -> false }t
t{ -tinf  1e    1.001e FABSDIFF< -> false }t

t{ +tinf  1e   +tinf   FABSDIFF< -> false }t
t{  1e   +tinf +tinf   FABSDIFF< -> false }t
t{ +tinf -tinf  1.001e FABSDIFF< -> false }t

t{ +tinf -tinf +tinf FABSDIFF< -> false }t

t{  0e    +tqnan +tinf FABSDIFF< -> false }t
t{ -tqnan  1e    +tinf FABSDIFF< -> false }t

?USE-NATIVE{
TESTING F~ exact tolerance

0 [IF]
\ BEGIN VFX
\ : FSIGN	\ F: fn -- |fn| ; -- flag ; true if negative
\   FSIGNBIT FABS ;

: F~            \ F: f1 f2 f3 -- ; -- flag
\ *G Approximation function. If f3 is positive, flag is true if abs[f1-f2]
\ ** less than f1. IF f3 is zero, flag is true if f1 and f2 have same
\ ** encoding. If f3 is negative, flag is true if abs[f1-f2] less than
\ ** abs[f3*abs[f1+f2]].
  fdup f0> if                           \ if f3 > 0
    frot frot  f- fabs			\ -- f3 |f1-f2|
    f>  exit                            \ -- t/f ; SFP007
  then
  fdup f0= if				\ if f3 = 0
    fdrop fsign >r
    fswap fsign >r
    f= r> r> xor 0= and  exit
  then
  frot frot  fover fover		\ -- f3 f1 f2 f1 f2
  f- fabs				\ -- f3 f1 f2 |f1-f2|
  frot frot				\ -- f3 |f1-f2| f1 f2
  fabs fswap fabs			\ -- f3 |f1-f2| |f1| |f2|
  f+ frot fabs f*			\ -- |f1-f2| |f3|*(|f1|+|f2|)
  f<					\ -- t/f
;

\  see F~
\ END VFX
[THEN]

t{ +t0 +t0 +t0 F~ -> true }t
t{ +t0 -t0 +t0 F~ -> false }t
t{ -t0 +t0 +t0 F~ -> false }t
t{ -t0 -t0 +t0 F~ -> true }t

t{  7e -2e +t0 F~ -> false }t
t{ -2e  7e +t0 F~ -> false }t
t{  7e  7e +t0 F~ -> true }t
   
t{  7e   +tinf +t0 F~ -> false }t
t{ +tinf  7e   +t0 F~ -> false }t
t{  7e   -tinf +t0 F~ -> false }t
t{ -tinf  7e   +t0 F~ -> false }t

t{ +tinf +tinf +t0 F~ -> true }t
t{ +tinf -tinf +t0 F~ -> false }t
t{ -tinf +tinf +t0 F~ -> false }t
t{ -tinf -tinf +t0 F~ -> true }t

t{ +tqnan  7e    +t0 F~ -> false }t
t{ -tqnan  7e    +t0 F~ -> false }t
t{  7e    +tqnan +t0 F~ -> false }t
t{  7e    -tqnan +t0 F~ -> false }t

t{ +tqnan +tqnan +t0 F~ -> true }t
t{ -tqnan +tqnan +t0 F~ -> false }t
t{ +tqnan -tqnan +t0 F~ -> false }t
t{ -tqnan -tqnan +t0 F~ -> true }t

t{ +tinf +tqnan +t0 F~ -> false }t
t{ -tinf +tqnan +t0 F~ -> false }t
t{ +tinf -tqnan +t0 F~ -> false }t
t{ -tinf -tqnan +t0 F~ -> false }t

t{ +tqnan +tinf +t0 F~ -> false }t
t{ -tqnan +tinf +t0 F~ -> false }t
t{ +tqnan -tinf +t0 F~ -> false }t
t{ -tqnan -tinf +t0 F~ -> false }t

t{ +t0 +t0 -t0 F~ -> true }t
t{ +t0 -t0 -t0 F~ -> false }t
t{ -t0 +t0 -t0 F~ -> false }t
t{ -t0 -t0 -t0 F~ -> true }t

t{  7e -2e -t0 F~ -> false }t
t{ -2e  7e -t0 F~ -> false }t
t{  7e  7e -t0 F~ -> true }t

t{  7e   +tinf -t0 F~ -> false }t
t{ +tinf  7e   -t0 F~ -> false }t
t{  7e   -tinf -t0 F~ -> false }t
t{ -tinf  7e   -t0 F~ -> false }t

t{ +tinf -tinf -t0 F~ -> false }t
t{ -tinf +tinf -t0 F~ -> false }t
t{ -tinf -tinf -t0 F~ -> true }t
t{ +tinf +tinf -t0 F~ -> true }t

t{ +tqnan  7e    -t0 F~ -> false }t
t{ -tqnan  7e    -t0 F~ -> false }t
t{  7e    +tqnan -t0 F~ -> false }t
t{  7e    -tqnan -t0 F~ -> false }t

t{ +tqnan +tqnan -t0 F~ -> true }t
t{ -tqnan +tqnan -t0 F~ -> false }t
t{ +tqnan -tqnan -t0 F~ -> false }t
t{ -tqnan -tqnan -t0 F~ -> true }t

t{ +tinf +tqnan -t0 F~ -> false }t
t{ -tinf +tqnan -t0 F~ -> false }t
t{ +tinf -tqnan -t0 F~ -> false }t
t{ -tinf -tqnan -t0 F~ -> false }t

t{ +tqnan +tinf -t0 F~ -> false }t
t{ -tqnan +tinf -t0 F~ -> false }t
t{ +tqnan -tinf -t0 F~ -> false }t
t{ -tqnan -tinf -t0 F~ -> false }t

TESTING F~ absolute tolerance

t{ +t0 +t0  7e F~ -> true }t
t{ +t0 -t0  7e F~ -> true }t
t{ -t0 +t0  7e F~ -> true }t
t{ -t0 -t0  7e F~ -> true }t

t{  7e   +tinf  7e F~ -> false }t
t{ +tinf  7e    7e F~ -> false }t
t{  7e   -tinf  7e F~ -> false }t
t{ -tinf  7e    7e F~ -> false }t

t{ +tinf +tinf  7e F~ -> false }t
t{ +tinf -tinf  7e F~ -> false }t
t{ -tinf +tinf  7e F~ -> false }t
t{ -tinf -tinf  7e F~ -> false }t

t{ +tqnan  7e    7e F~ -> false }t
t{ -tqnan  7e    7e F~ -> false }t
t{  7e    +tqnan 7e F~ -> false }t
t{  7e    -tqnan 7e F~ -> false }t

t{ +tqnan +tqnan  7e F~ -> false }t
t{ -tqnan +tqnan  7e F~ -> false }t
t{ +tqnan -tqnan  7e F~ -> false }t
t{ -tqnan -tqnan  7e F~ -> false }t

t{ +tinf +tqnan  7e F~ -> false }t
t{ -tinf +tqnan  7e F~ -> false }t
t{ +tinf -tqnan  7e F~ -> false }t
t{ -tinf -tqnan  7e F~ -> false }t

t{ +tqnan +tinf  7e F~ -> false }t
t{ -tqnan +tinf  7e F~ -> false }t
t{ +tqnan -tinf  7e F~ -> false }t
t{ -tqnan -tinf  7e F~ -> false }t

t{ +t0 +t0 +tinf F~ -> true }t
t{ +t0 -t0 +tinf F~ -> true }t
t{ -t0 +t0 +tinf F~ -> true }t
t{ -t0 -t0 +tinf F~ -> true }t

t{  7e -2e +tinf F~ -> true }t
t{ -2e  7e +tinf F~ -> true }t
t{  7e  7e +tinf F~ -> true }t

t{  7e   +tinf +tinf F~ -> false }t
t{ +tinf  7e   +tinf F~ -> false }t
t{  7e   -tinf +tinf F~ -> false }t
t{ -tinf  7e   +tinf F~ -> false }t

t{ +tinf +tinf +tinf F~ -> false }t
t{ +tinf -tinf +tinf F~ -> false }t
t{ -tinf +tinf +tinf F~ -> false }t
t{ -tinf -tinf +tinf F~ -> false }t

t{ +tqnan  7e    +tinf F~ -> false }t
t{ -tqnan  7e    +tinf F~ -> false }t
t{  7e    +tqnan +tinf F~ -> false }t
t{  7e    -tqnan +tinf F~ -> false }t

t{ +tqnan +tqnan +tinf F~ -> false }t
t{ -tqnan +tqnan +tinf F~ -> false }t
t{ +tqnan -tqnan +tinf F~ -> false }t
t{ -tqnan -tqnan +tinf F~ -> false }t

t{ +tinf +tqnan +tinf F~ -> false }t
t{ -tinf +tqnan +tinf F~ -> false }t
t{ +tinf -tqnan +tinf F~ -> false }t
t{ -tinf -tqnan +tinf F~ -> false }t

t{ +tqnan +tinf +tinf F~ -> false }t
t{ -tqnan +tinf +tinf F~ -> false }t
t{ +tqnan -tinf +tinf F~ -> false }t
t{ -tqnan -tinf +tinf F~ -> false }t

TESTING F~ relative tolerance

t{ +t0 +t0 -7e F~ -> false }t
t{ +t0 -t0 -7e F~ -> false }t
t{ -t0 +t0 -7e F~ -> false }t
t{ -t0 -t0 -7e F~ -> false }t

t{  7e   +tinf -7e F~ -> false }t
t{ +tinf  7e   -7e F~ -> false }t
t{  7e   -tinf -7e F~ -> false }t
t{ -tinf  7e   -7e F~ -> false }t

t{ +tinf +tinf -7e F~ -> false }t
t{ +tinf -tinf -7e F~ -> false }t
t{ -tinf +tinf -7e F~ -> false }t
t{ -tinf -tinf -7e F~ -> false }t

t{ +tqnan  7e    -7e F~ -> false }t
t{ -tqnan  7e    -7e F~ -> false }t
t{  7e    +tqnan -7e F~ -> false }t
t{  7e    -tqnan -7e F~ -> false }t

t{ +tqnan +tqnan -7e F~ -> false }t
t{ -tqnan +tqnan -7e F~ -> false }t
t{ +tqnan -tqnan -7e F~ -> false }t
t{ -tqnan -tqnan -7e F~ -> false }t

t{ +tinf +tqnan -7e F~ -> false }t
t{ -tinf +tqnan -7e F~ -> false }t
t{ +tinf -tqnan -7e F~ -> false }t
t{ -tinf -tqnan -7e F~ -> false }t

t{ +tqnan +tinf -7e F~ -> false }t
t{ -tqnan +tinf -7e F~ -> false }t
t{ +tqnan -tinf -7e F~ -> false }t
t{ -tqnan -tinf -7e F~ -> false }t

t{ +t0 +t0 -tinf F~ -> false }t
t{ +t0 -t0 -tinf F~ -> false }t
t{ -t0 +t0 -tinf F~ -> false }t
t{ -t0 -t0 -tinf F~ -> false }t

t{  7e -2e -tinf F~ -> true }t
t{ -2e  7e -tinf F~ -> true }t
t{  7e  7e -tinf F~ -> true }t

t{  7e   +tinf -tinf F~ -> false }t
t{ +tinf 7e    -tinf F~ -> false }t
t{  7e  -tinf  -tinf F~ -> false }t
t{ -tinf 7e    -tinf F~ -> false }t

t{ +tinf +tinf -tinf F~ -> false }t
t{ +tinf -tinf -tinf F~ -> false }t
t{ -tinf +tinf -tinf F~ -> false }t
t{ -tinf -tinf -tinf F~ -> false }t

t{ +tqnan  7e    -tinf F~ -> false }t
t{ -tqnan  7e    -tinf F~ -> false }t
t{  7e    +tqnan -tinf F~ -> false }t
t{  7e    -tqnan -tinf F~ -> false }t

t{ +tqnan +tqnan -tinf F~ -> false }t
t{ -tqnan +tqnan -tinf F~ -> false }t
t{ +tqnan -tqnan -tinf F~ -> false }t
t{ -tqnan -tqnan -tinf F~ -> false }t

t{ +tinf +tqnan -tinf F~ -> false }t
t{ -tinf +tqnan -tinf F~ -> false }t
t{ +tinf -tqnan -tinf F~ -> false }t
t{ -tinf -tqnan -tinf F~ -> false }t

t{ +tqnan +tinf -tinf F~ -> false }t
t{ -tqnan +tinf -tinf F~ -> false }t
t{ +tqnan -tinf -tinf F~ -> false }t
t{ -tqnan -tinf -tinf F~ -> false }t

TESTING F~ nan tolerance

t{ +t0 +t0 +tqnan F~ -> false }t
t{ +t0 -t0 +tqnan F~ -> false }t
t{ -t0 +t0 +tqnan F~ -> false }t
t{ -t0 -t0 +tqnan F~ -> false }t

t{  7e -2e +tqnan F~ -> false }t
t{ -2e  7e +tqnan F~ -> false }t
t{  7e  7e +tqnan F~ -> false }t

t{  7e   +tinf +tqnan F~ -> false }t
t{ +tinf  7e   +tqnan F~ -> false }t
t{  7e   -tinf +tqnan F~ -> false }t
t{ -tinf   7e  +tqnan F~ -> false }t

t{ +tinf +tinf +tqnan F~ -> false }t
t{ +tinf -tinf +tqnan F~ -> false }t
t{ -tinf +tinf +tqnan F~ -> false }t
t{ -tinf -tinf +tqnan F~ -> false }t

t{ +tqnan  7e    +tqnan F~ -> false }t
t{ -tqnan  7e    +tqnan F~ -> false }t
t{  7e    +tqnan +tqnan F~ -> false }t
t{  7e    -tqnan +tqnan F~ -> false }t

t{ +tqnan +tqnan +tqnan F~ -> false }t
t{ -tqnan +tqnan +tqnan F~ -> false }t
t{ +tqnan -tqnan +tqnan F~ -> false }t
t{ -tqnan -tqnan +tqnan F~ -> false }t

t{ +tinf +tqnan +tqnan F~ -> false }t
t{ -tinf +tqnan +tqnan F~ -> false }t
t{ +tinf -tqnan +tqnan F~ -> false }t
t{ -tinf -tqnan +tqnan F~ -> false }t

t{ +tqnan +tinf +tqnan F~ -> false }t
t{ -tqnan +tinf +tqnan F~ -> false }t
t{ +tqnan -tinf +tqnan F~ -> false }t
t{ -tqnan -tinf +tqnan F~ -> false }t

t{ +t0 +t0 +tqnan F~ -> false }t
t{ +t0 -t0 +tqnan F~ -> false }t
t{ -t0 +t0 +tqnan F~ -> false }t
t{ -t0 -t0 +tqnan F~ -> false }t

t{  7e -2e -tqnan F~ -> false }t
t{ -2e  7e -tqnan F~ -> false }t
t{  7e  7e -tqnan F~ -> false }t

t{  7e   +tinf -tqnan F~ -> false }t
t{ +tinf  7e   -tqnan F~ -> false }t
t{  7e   -tinf -tqnan F~ -> false }t
t{ -tinf  7e   -tqnan F~ -> false }t

t{ +tinf +tinf -tqnan F~ -> false }t
t{ +tinf -tinf -tqnan F~ -> false }t
t{ -tinf +tinf -tqnan F~ -> false }t
t{ -tinf -tinf -tqnan F~ -> false }t

t{ +tqnan  7e    -tqnan F~ -> false }t
t{ -tqnan  7e    -tqnan F~ -> false }t
t{  7e    +tqnan -tqnan F~ -> false }t
t{  7e    -tqnan -tqnan F~ -> false }t

t{ +tqnan +tqnan -tqnan F~ -> false }t
t{ -tqnan +tqnan -tqnan F~ -> false }t
t{ +tqnan -tqnan -tqnan F~ -> false }t
t{ -tqnan -tqnan -tqnan F~ -> false }t

t{ +tinf +tqnan -tqnan F~ -> false }t
t{ -tinf +tqnan -tqnan F~ -> false }t
t{ +tinf -tqnan -tqnan F~ -> false }t
t{ -tinf -tqnan -tqnan F~ -> false }t

t{ +tqnan +tinf -tqnan F~ -> false }t
t{ -tqnan +tinf -tqnan F~ -> false }t
t{ +tqnan -tinf -tqnan F~ -> false }t
t{ -tqnan -tinf -tqnan F~ -> false }t

TESTING FMAX FMIN

t{  7e    7e   FMAX ->  7e }t
t{  7e    7.1e FMAX ->  7.1e }t
t{  7.1e  7e   FMAX ->  7.1e }t

t{  7e   +tinf FMAX -> +tinf }t
t{ +tinf  7e   FMAX -> +tinf }t
t{  7e   -tinf FMAX ->  7e }t
t{ -tinf  7e   FMAX ->  7e }t

t{ +tinf +tinf FMAX -> +tinf }t
t{ +tinf -tinf FMAX -> +tinf }t
t{ -tinf +tinf FMAX -> +tinf }t
t{ -tinf -tinf FMAX -> -tinf }t

t{  7e    -tqnan FMAX ->  7e }t
t{ -tqnan  7e    FMAX ->  7e }t
t{ -7e    +tqnan FMAX -> -7e }t
t{ +tqnan -7e    FMAX -> -7e }t

t{ +t0    -tqnan FMAX -> +t0 }t
t{ -tqnan +t0    FMAX -> +t0 }t
t{ -t0    +tqnan FMAX -> -t0 }t
t{ +tqnan -t0    FMAX -> -t0 }t

\ IEEE 754-2008 allows either nan to be returned.
t{ +tqnan +tqnan FMAX      -> +tqnan }t
t{ +tqnan -tqnan FMAX FABS -> +tqnan }t
t{ -tqnan +tqnan FMAX FABS -> +tqnan }t
t{ -tqnan -tqnan FMAX      -> -tqnan }t

\ IEEE 754-2008 allows either signed zero to be returned.
t{ -t0 -t0 FMAX      -> -t0 }t
t{ -t0 +t0 FMAX FABS -> +t0 }t
t{ +t0 -t0 FMAX FABS -> +t0 }t
t{ +t0 +t0 FMAX      -> +t0 }t

t{  7e    7e   FMIN ->  7e }t
t{  7e    7.1e FMIN ->  7e }t
t{  7.1e  7e   FMIN ->  7e }t

t{  7e   +tinf FMIN ->  7e }t
t{ +tinf  7e   FMIN ->  7e }t
t{  7e   -tinf FMIN -> -tinf }t
t{ -tinf  7e   FMIN -> -tinf }t

t{ +tinf +tinf FMIN -> +tinf }t
t{ +tinf -tinf FMIN -> -tinf }t
t{ -tinf +tinf FMIN -> -tinf }t
t{ -tinf -tinf FMIN -> -tinf }t

t{  7e    -tqnan FMIN ->  7e }t
t{ -tqnan  7e    FMIN ->  7e }t
t{ -7e    +tqnan FMIN -> -7e }t
t{ +tqnan -7e    FMIN -> -7e }t

t{ +t0    -tqnan FMIN -> +t0 }t
t{ -tqnan +t0    FMIN -> +t0 }t
t{ -t0    +tqnan FMIN -> -t0 }t
t{ +tqnan -t0    FMIN -> -t0 }t

\ IEEE 754-2008 allows either nan to be returned.
t{ +tqnan +tqnan FMIN      -> +tqnan }t
t{ +tqnan -tqnan FMIN FABS -> +tqnan }t
t{ -tqnan +tqnan FMIN FABS -> +tqnan }t
t{ -tqnan -tqnan FMIN      -> -tqnan }t

\ IEEE 754-2008 allows either signed zero to be returned.
t{ -t0 -t0 FMIN      -> -t0 }t
t{ -t0 +t0 FMIN FABS -> +t0 }t
t{ +t0 -t0 FMIN FABS -> +t0 }t
t{ +t0 +t0 FMIN      -> +t0 }t
}?USE-NATIVE

testing FNEXTUP FNEXTDOWN

bits/float #32 = [IF]
  $40000001 raw>f FCONSTANT [+2]-up
  $3fffffff raw>f FCONSTANT [+2]-down
  $bfffffff raw>f FCONSTANT [-2]-up
  $c0000001 raw>f FCONSTANT [-2]-down
  $7f7ffffe raw>f FCONSTANT [+tmax-n]-down
  $ff7ffffe raw>f FCONSTANT [-tmax-n]-up
  $00800001 raw>f FCONSTANT [+tmin-n]-up
  $80800001 raw>f FCONSTANT [-tmin-n]-down [THEN]

bits/cell #32 = [IF]
bits/float #64 =  [IF]
  $00000001 $40000000 raw>f FCONSTANT [+2]-up
  $ffffffff $3fffffff raw>f FCONSTANT [+2]-down
  $ffffffff $bfffffff raw>f FCONSTANT [-2]-up
  $00000001 $c0000000 raw>f FCONSTANT [-2]-down
  $fffffffe $7fefffff raw>f FCONSTANT [+tmax-n]-down
  $fffffffe $ffefffff raw>f FCONSTANT [-tmax-n]-up
  $00000001 $00100000 raw>f FCONSTANT [+tmin-n]-up
  $00000001 $80100000 raw>f FCONSTANT [-tmin-n]-down [THEN]
bits/float #80 =  [IF]
  $00000001 $80000000 $4000 raw>f FCONSTANT [+2]-up
  $ffffffff $ffffffff $3fff raw>f FCONSTANT [+2]-down
  $ffffffff $ffffffff $bfff raw>f FCONSTANT [-2]-up
  $00000001 $80000000 $c000 raw>f FCONSTANT [-2]-down
  $fffffffe $ffffffff $7ffe raw>f FCONSTANT [+tmax-n]-down
  $fffffffe $ffffffff $fffe raw>f FCONSTANT [-tmax-n]-up
  $00000001 $80000000 $0001 raw>f FCONSTANT [+tmin-n]-up
  $00000001 $80000000 $8001 raw>f FCONSTANT [-tmin-n]-down [THEN]
bits/float #128 = [IF] cr .( ***Missing quad FCONSTANTs.) cr [THEN]

[ELSE]  \ 64 bits/cell
bits/float #64 =  [IF]
  $4000000000000001 raw>f FCONSTANT [+2]-up
  $3fffffffffffffff raw>f FCONSTANT [+2]-down
  $bfffffffffffffff raw>f FCONSTANT [-2]-up
  $c000000000000001 raw>f FCONSTANT [-2]-down
  $7feffffffffffffe raw>f FCONSTANT [+tmax-n]-down
  $ffeffffffffffffe raw>f FCONSTANT [-tmax-n]-up
  $0010000000000001 raw>f FCONSTANT [+tmin-n]-up
  $8010000000000001 raw>f FCONSTANT [-tmin-n]-down [THEN]
bits/float #80 =  [IF]
  $8000000000000001 $4000 raw>f FCONSTANT [+2]-up
  $ffffffffffffffff $3fff raw>f FCONSTANT [+2]-down
  $ffffffffffffffff $bfff raw>f FCONSTANT [-2]-up
  $8000000000000001 $c000 raw>f FCONSTANT [-2]-down
  $fffffffffffffffe $7ffe raw>f FCONSTANT [+tmax-n]-down
  $fffffffffffffffe $fffe raw>f FCONSTANT [-tmax-n]-up
  $8000000000000001 $0001 raw>f FCONSTANT [+tmin-n]-up
  $8000000000000001 $8001 raw>f FCONSTANT [-tmin-n]-down [THEN]
bits/float #128 = [IF] cr .( ***Missing quad FCONSTANTs.) cr [THEN]
[THEN]

: RAW-QUIETED  ( nan.raw -- qnan.raw )
(
Assume the input to be a raw nan.  Leave the raw nan with quiet bit on
and all other bits unchanged [quieted simply].
)
  [ bits/cell #32 = ] [IF]
    [ bits/float #32 = ]  [IF] $0040000  or           [THEN]
    [ bits/float #64 = ]  [IF] $00080000 or           [THEN]
    [ bits/float #80 = ]  [IF] swap $40000000 or swap [THEN]
    [ bits/float #128 = ] [IF] $00008000 or [THEN]
  [ELSE]  \ bits/cell = 64
    [ bits/float #32 = ]  [IF] $00400000 or                   [THEN]
    [ bits/float #64 = ]  [IF] $0008000000000000 or           [THEN]
    [ bits/float #80 = ]  [IF] swap $4000000000000000 or swap [THEN]
    [ bits/float #128 = ] [IF] $0000800000000000 or           [THEN]
  [THEN] ;

: QUIETED  ( snan|qnan -- qnan )  f>raw raw-quieted raw>f ;

\ Passing these tests supports the conjecture that F+ with one nan and
\ one extended number input produces the nan quieted simply.
t{ +max-snan-raw raw>f  0.0e f+ f>raw -> +max-snan-raw raw-quieted }t
t{ +min-snan-raw raw>f  0.0e f+ f>raw -> +min-snan-raw raw-quieted }t
t{ -max-snan-raw raw>f  0.0e f+ f>raw -> -max-snan-raw raw-quieted }t
t{ -min-snan-raw raw>f  0.0e f+ f>raw -> -min-snan-raw raw-quieted }t
t{ +max-qnan-raw raw>f  0.0e f+ f>raw -> +max-qnan-raw }t
t{ +min-qnan-raw raw>f  0.0e f+ f>raw -> +min-qnan-raw }t
t{ -max-qnan-raw raw>f  0.0e f+ f>raw -> -max-qnan-raw }t
t{ -min-qnan-raw raw>f  0.0e f+ f>raw -> -min-qnan-raw }t
t{ +max-snan-raw raw>f +tinf f+ f>raw -> +max-snan-raw raw-quieted }t
t{ +max-snan-raw raw>f -tinf f+ f>raw -> +max-snan-raw raw-quieted }t

\ qnan propagation
t{ +tqnan FNEXTUP FQNAN? -> true }t
t{ -tqnan FNEXTUP FQNAN? -> true }t
\ If these pass, these particular qNaNs are conserved.
t{ +tqnan FNEXTUP -> +tqnan }t \ not an ieee "shall"
t{ -tqnan FNEXTUP -> -tqnan }t \ not an ieee "shall"

\ snan quieting
t{ +tsnan FNEXTUP FQNAN? -> true }t
t{ -tsnan FNEXTUP FQNAN? -> true }t
\ If these pass, these particular sNaNs are quieted simply.
t{ +tsnan FNEXTUP -> +tsnan quieted }t \ not an ieee "shall"
t{ -tsnan FNEXTUP -> -tsnan quieted }t \ not an ieee "shall"

\ +max-n boundary
t{ +tinf          FNEXTUP -> +tinf }t
t{ +tmax-n        FNEXTUP -> +tinf }t
t{ [+tmax-n]-down FNEXTUP -> +tmax-n }t

\ +normal integer boundaries
t{  2e        FNEXTUP -> [+2]-up }t
t{  [+2]-down FNEXTUP ->  2e }t

\ +subnormal boundaries
t{ +tmin-n    FNEXTUP -> [+tmin-n]-up }t
t{ +tmax-subn FNEXTUP -> +tmin-n }t
t{ +tmin-subn FNEXTUP -> +tmin-subn f2* }t

\ +|-0 boundaries
t{ +t0        FNEXTUP -> +tmin-subn }t
t{ -t0        FNEXTUP -> +tmin-subn }t
t{ -tmin-subn FNEXTUP -> -t0 }t

\ -subnormal boundaries
t{ -tmin-subn f2* FNEXTUP -> -tmin-subn }t
t{ -tmin-n        FNEXTUP -> -tmax-subn }t
t{ [-tmin-n]-down FNEXTUP -> -tmin-n }t

\ -normal integer boundaries
t{ [-2]-down FNEXTUP -> -2e }t
t{ -2e       FNEXTUP -> [-2]-up }t 

\ -max-n boundary
t{ -tmax-n FNEXTUP -> [-tmax-n]-up }t
t{ -tinf   FNEXTUP -> -tmax-n }t

\ tnan propagation
t{ +tqnan FNEXTDOWN FQNAN? -> true }t
t{ -tqnan FNEXTDOWN FQNAN? -> true }t
\ If these pass, these particular qNaNs are conserved.
t{ +tqnan FNEXTDOWN -> +tqnan }t \ not an ieee "shall"
t{ -tqnan FNEXTDOWN -> -tqnan }t \ not an ieee "shall"

\ snan quieting
t{ +tsnan FNEXTDOWN FQNAN? -> true }t
t{ -tsnan FNEXTDOWN FQNAN? -> true }t
\ If these pass, these particular sNaNs are quieted simply.
t{ +tsnan FNEXTDOWN -> +tsnan quieted }t \ not an ieee "shall"
t{ -tsnan FNEXTDOWN -> -tsnan quieted }t \ not an ieee "shall"

\ max-n boundary
t{ +tinf   FNEXTDOWN -> +tmax-n }t
t{ +tmax-n FNEXTDOWN -> [+tmax-n]-down }t

\ +normal integer boundaries
t{ [+2]-up FNEXTDOWN ->  2e }t
t{  2e     FNEXTDOWN -> [+2]-down }t

\ +subnormal boundaries
t{ [+tmin-n]-up   FNEXTDOWN -> +tmin-n }t
t{ +tmin-n        FNEXTDOWN -> +tmax-subn }t 
t{ +tmin-subn f2* FNEXTDOWN -> +tmin-subn }t 

\ +|-0 boundaries
t{ +tmin-subn FNEXTDOWN -> +t0 }t
t{ +t0        FNEXTDOWN -> -tmin-subn }t
t{ -t0        FNEXTDOWN -> -tmin-subn }t

\ -subnormal boundaries
t{ -tmin-subn FNEXTDOWN -> -tmin-subn f2* }t
t{ -tmax-subn FNEXTDOWN -> -tmin-n }t
t{ -tmin-n    FNEXTDOWN -> [-tmin-n]-down }t

\ -normal integer boundaries
t{ [-2]-up FNEXTDOWN -> -2e }t 
t{ -2e     FNEXTDOWN -> [-2]-down }t 

\ -max-n boundary
t{ [-tmax-n]-up FNEXTDOWN -> -tmax-n }t
t{ -tmax-n      FNEXTDOWN -> -tinf }t
t{ -tinf        FNEXTDOWN -> -tinf }t

?.errors GFORTH-HOST [IF] ?.cr [THEN]
